@@ -3,6 +3,7 @@
 import enum
 import itertools
 import logging
+import mimetypes
 import os
 import random
 from re import sub
@@ -15,6 +16,7 @@ from parameterized import parameterized
 import requests
 
 from joppy.api import Api
+from joppy import tools
 from . import setup_joplin
 
 
@@ -27,6 +29,7 @@ logging.basicConfig(
 
 
 SLOW_TESTS = bool(os.getenv("SLOW_TESTS", ""))
+PROFILE = "test_profile"
 API_TOKEN = os.getenv("API_TOKEN", "")
 APP = None
 
@@ -45,15 +48,6 @@ def with_resource(func):
     return inner_decorator
 
 
-def powerset(iterable):
-    """powerset([1,2,3]) --> (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"""
-    # https://docs.python.org/2/library/itertools.html#recipes
-    list_ = list(iterable)
-    return itertools.chain.from_iterable(
-        itertools.combinations(list_, r) for r in range(1, len(list_) + 1)
-    )
-
-
 class ChangeType(enum.Enum):
     # https://joplinapp.org/api/references/rest_api/#properties-4
     CREATED = 1
@@ -88,8 +82,8 @@ def setUpModule():
     if not API_TOKEN:
         app_path = "./joplin.AppImage"
         setup_joplin.download_joplin(app_path)
-        APP = setup_joplin.JoplinApp(app_path)
-        API_TOKEN = APP.get_token()
+        APP = setup_joplin.JoplinApp(app_path, profile=PROFILE)
+        API_TOKEN = APP.api_token
 
 
 def tearDownModule():
@@ -107,6 +101,7 @@ class TestBase(unittest.TestCase):
         self.api.delete_all_notebooks()
         self.api.delete_all_resources()
         self.api.delete_all_tags()
+        self.current_cursor = self.api.get_events()["cursor"]
 
     @staticmethod
     def get_random_id() -> str:
@@ -144,13 +139,20 @@ class TestBase(unittest.TestCase):
         return 0 <= timestamp <= int(time.time() * 1000)  # ms
 
     @staticmethod
-    def get_combinations(iterable):
-        """Get combinations of an iterable."""
-        combinations = powerset(iterable)
-        if not SLOW_TESTS:
-            # Only chose some combinations.
-            combinations = random.choices(list(combinations), k=50)
-        return combinations
+    def get_combinations(iterable, max_combinations: int = 100):
+        """Get some combinations of an iterable."""
+        # https://stackoverflow.com/a/10465588
+        # TODO: Randomize fully. For now the combinations are sorted by length.
+        list_ = list(iterable)
+        lengths = list(range(1, len(list_) + 1))
+        random.shuffle(lengths)
+        combinations = itertools.chain.from_iterable(
+            itertools.combinations(list_, r)
+            for r in lengths
+            # shuffle each iteration
+            if random.shuffle(list_) is None  # type: ignore
+        )
+        return itertools.islice(combinations, max_combinations)
 
 
 class Event(TestBase):
@@ -162,17 +164,31 @@ class Event(TestBase):
     ]
     default_properties = ["id", "item_type", "item_id", "type", "created_time"]
 
+    def generate_event(self):
+        """Generate an event and wait until it's available."""
+        events_before = len(self.api.get_all_events(cursor=self.current_cursor))
+        # For now only notes trigger events:
+        # https://joplinapp.org/api/references/rest_api/#events
+        self.api.add_notebook()
+        self.api.add_note()
+
+        def compare_event_count():
+            event_count = len(self.api.get_all_events(cursor=self.current_cursor))
+            return event_count if event_count == events_before + 1 else None
+
+        # Wait until the event is available.
+        setup_joplin.wait_for(compare_event_count, interval=0.1, timeout=1)
+
     def test_get_event(self):
         """Get a specific event."""
-        events = self.api.get_events(cursor=0)
-        random_event = random.choice(events["items"])
-        event = self.api.get_event(id_=random_event["id"])
+        self.generate_event()
+        last_event = self.api.get_events(cursor=self.current_cursor)["items"][-1]
+        event = self.api.get_event(id_=last_event["id"])
 
         self.assertEqual(list(event.keys()), self.default_properties + ["type_"])
         self.assertIn(event["item_type"], ItemType._value2member_map_)
         self.assertTrue(self.is_id_valid(event["item_id"]))
-        # Because of the cleanup before the test.
-        self.assertEqual(event["type"], ChangeType.DELETED.value)
+        self.assertEqual(event["type"], ChangeType.CREATED.value)
         self.assertTrue(self.is_timestamp_valid(event["created_time"]))
         self.assertEqual(event["type_"], ItemType.ITEM_CHANGE.value)
 
@@ -181,9 +197,8 @@ class Event(TestBase):
 
     def test_get_events_by_cursor(self):
         """Get all events by specifying a cursor of 0."""
-        # Events should be filled by previous API calls.
+        self.generate_event()
         events = self.api.get_events(cursor=0)
-        self.assertGreater(len(events["items"]), 0)
 
         previous_created_time = 0
         previous_id = 0
@@ -222,8 +237,7 @@ class Note(TestBase):
         "application_data", "order", "user_created_time", "user_updated_time",
         "encryption_cipher_text", "encryption_applied", "markup_language",
         "is_shared", "share_id", "conflict_original_id",
-        # "body_html", "base_url",
-        # "image_data_url", "crop_rect",
+        # "body_html", "base_url", "image_data_url", "crop_rect",
         # fmt: on
     ]
     default_properties = ["id", "parent_id", "title"]
@@ -241,26 +255,23 @@ class Note(TestBase):
         self.assertEqual(notes[0]["id"], id_)
         self.assertEqual(notes[0]["parent_id"], parent_id)
 
-    def test_add_all_properties(self):
-        """Add a note and define all user definable properties."""
+    def test_add_attach_image(self):
+        """Add a note with an image attached."""
         self.api.add_notebook()
-        id_ = self.get_random_id()
-        title = "aaa"
-        user_created_time = 0  # positive integer
-        user_updated_time = 1  # positive integer
-        is_shared = int(True)  # 0 (False) or 1 (True)
-        share_id = 1
-        self.api.add_note(
-            id_=id_,
-            title=title,
-            user_created_time=user_created_time,
-            user_updated_time=user_updated_time,
-            is_shared=is_shared,
-            share_id=share_id,
+        image_data = tools.encode_base64("test/grant_authorization_button.png")
+        note_id = self.api.add_note(
+            image_data_url=f"data:image/png;base64,{image_data}"
         )
 
-        notes = self.api.get_notes(fields=",".join(self.properties))["items"]
-        print(notes)
+        # Check that there is a resource.
+        resources = self.api.get_resources()["items"]
+        self.assertEqual(len(resources), 1)
+        resource_id = resources[0]["id"]
+
+        # Verify the resource is attached to the note.
+        resources = self.api.get_resources(note_id=note_id)["items"]
+        self.assertEqual(len(resources), 1)
+        self.assertEqual(resources[0]["id"], resource_id)
 
     def test_add_no_notebook(self):
         """A note has to be added to an existing notebook."""
@@ -272,13 +283,6 @@ class Note(TestBase):
             context.exception.response.json()["error"],
         )
 
-    def test_add_long_title(self):
-        """Add a note with a long title."""
-        title = self.get_random_string(10 ** 6)
-        self.api.add_notebook()
-        note_id = self.api.add_note(title=title)
-        self.assertEqual(self.api.get_note(id_=note_id)["title"], title)
-
     def test_delete(self):
         """Add and then delete a note."""
         self.api.add_notebook()
@@ -307,6 +311,18 @@ class Note(TestBase):
         for note in notes["items"]:
             self.assertEqual(list(note.keys()), self.default_properties)
 
+    def test_get_all_notes(self):
+        """Get all notes, unpaginated."""
+        self.api.add_notebook()
+        # Small limit and count to create/remove as less as possible items.
+        count, limit = random.randint(1, 10), random.randint(1, 10)
+        for _ in range(count):
+            self.api.add_note()
+        self.assertEqual(
+            len(self.api.get_notes(limit=limit)["items"]), min(limit, count)
+        )
+        self.assertEqual(len(self.api.get_all_notes(limit=limit)), count)
+
     def test_get_notes_too_many_ids(self):
         """At maximum one parent ID can be used to obtain notes."""
         with self.assertRaises(ValueError):
@@ -314,8 +330,6 @@ class Note(TestBase):
 
     def test_get_notes_valid_properties(self):
         """Try to get specific properties of a note."""
-        self.skipTest("TODO: Too many combinations. RAM is fully used.")
-
         self.api.add_notebook()
         self.api.add_note()
         property_combinations = self.get_combinations(self.properties)
@@ -358,33 +372,6 @@ class Notebook(TestBase):
         self.assertEqual(len(notebooks), 1)
         self.assertEqual(notebooks[0]["id"], id_)
 
-    def test_add_all_properties(self):
-        """Add a notebook and define all user definable properties."""
-        id_ = self.get_random_id()
-        title = "aaa"
-        user_created_time = 1
-        user_updated_time = 2
-        is_shared = int(True)
-        share_id = 1
-        self.api.add_notebook(
-            id_=id_,
-            title=title,
-            user_created_time=user_created_time,
-            user_updated_time=user_updated_time,
-            is_shared=is_shared,
-            share_id=share_id,
-        )
-
-        notebooks = self.api.get_notebooks(fields=",".join(self.properties))["items"]
-        print(notebooks)
-
-    def test_add_sub_notebook(self):
-        """Add a sub notebook."""
-        id_ = self.api.add_notebook()
-        self.api.add_notebook(parent_id=id_)
-        notebooks = self.api.get_notebooks(fields=",".join(self.properties))["items"]
-        print(notebooks)
-
     def test_get_notebook(self):
         """Get a specific notebook."""
         id_ = self.api.add_notebook()
@@ -402,6 +389,17 @@ class Notebook(TestBase):
         for notebook in notebooks["items"]:
             self.assertEqual(list(notebook.keys()), self.default_properties)
 
+    def test_get_all_notebooks(self):
+        """Get all notebooks, unpaginated."""
+        # Small limit and count to create/remove as less as possible items.
+        count, limit = random.randint(1, 10), random.randint(1, 10)
+        for _ in range(count):
+            self.api.add_notebook()
+        self.assertEqual(
+            len(self.api.get_notebooks(limit=limit)["items"]), min(limit, count)
+        )
+        self.assertEqual(len(self.api.get_all_notebooks(limit=limit)), count)
+
     def test_get_notebooks_invalid_property(self):
         """Try to get an non existent property of notebooks."""
         self.api.add_notebook()
@@ -471,20 +469,23 @@ class Resource(TestBase):
         self.assertEqual(len(resources), 1)
         self.assertEqual(resources[0]["id"], id_)
 
-    # @with_resource
-    # def test_add_all_properties(self, filename: str):
-    #     """Add a resource with all user definable properties."""
-    #     id_ = self.get_random_id()
-    #     mime = "aaa"
-    #     user_created_time = 0
-    #     user_updated_time = 1
-    #     file_extension = "abc"
+    @with_resource
+    def test_add_to_note(self, filename: str):
+        """Add a resource to an existing note."""
+        self.api.add_notebook()
+        note_id = self.api.add_note()
+        resource_id = self.api.add_resource(filename=filename)
+        self.api.add_resource_to_note(resource_id=resource_id, note_id=note_id)
 
-    #     print(filename)
-    #     self.api.add_resource(id_=id_, filename=filename, mime=mime, user_created_time=user_created_time, user_updated_time=user_updated_time, file_extension=file_extension)
+        # Verify the resource is attached to the note.
+        resources = self.api.get_resources(note_id=note_id)["items"]
+        self.assertEqual(len(resources), 1)
+        self.assertEqual(resources[0]["id"], resource_id)
 
-    #     resources = self.api.get_resources(fields=",".join(self.properties))["items"]
-    #     print(resources)
+        # TODO: Seems to be not working.
+        # notes = self.api.get_notes(resource_id=resource_id)["items"]
+        # self.assertEqual(len(notes), 1)
+        # self.assertEqual(notes[0]["id"], note_id)
 
     @with_resource
     def test_delete(self, filename: str):
@@ -495,8 +496,7 @@ class Resource(TestBase):
 
         self.api.delete_resource(id_=id_)
         self.assertEqual(self.api.get_resources(), self.empty_search)
-        # TODO: How to obtain profile?
-        self.assertEqual(os.listdir("test_profile/resources"), [])
+        self.assertEqual(os.listdir(f"{PROFILE}/resources"), [])
 
     @with_resource
     def test_get_resource(self, filename):
@@ -516,6 +516,18 @@ class Resource(TestBase):
         for resource in resources["items"]:
             self.assertEqual(list(resource.keys()), self.default_properties)
 
+    @with_resource
+    def test_get_all_resources(self, filename):
+        """Get all resources, unpaginated."""
+        # Small limit and count to create/remove as less as possible items.
+        count, limit = random.randint(1, 10), random.randint(1, 10)
+        for _ in range(count):
+            self.api.add_resource(filename=filename)
+        self.assertEqual(
+            len(self.api.get_resources(limit=limit)["items"]), min(limit, count)
+        )
+        self.assertEqual(len(self.api.get_all_resources(limit=limit)), count)
+
     @with_resource
     def test_get_resources_valid_properties(self, filename):
         """Try to get specific properties of a resource."""
@@ -536,20 +548,26 @@ class Resource(TestBase):
         self.assertEqual(self.api.get_resource(id_=id_)["title"], new_title)
 
     @with_resource
-    def test_check_property_size(self, filename):
-        """Check the size of a resource."""
+    def test_check_derived_properties(self, filename):
+        """Check the derived properties. I. e. mime type, extension and size."""
         for file_ in ["test/grant_authorization_button.png", filename]:
             id_ = self.api.add_resource(filename=file_)
-            resource = self.api.get_resource(id_=id_, fields="size")
+            resource = self.api.get_resource(id_=id_, fields="mime,file_extension,size")
+            mime_type, _ = mimetypes.guess_type(file_)
+            self.assertEqual(
+                resource["mime"],
+                mime_type if mime_type is not None else "application/octet-stream",
+            )
+            self.assertEqual(resource["file_extension"], os.path.splitext(file_)[1][1:])
             self.assertEqual(resource["size"], os.path.getsize(file_))
 
     @with_resource
-    def test_check_property_extension(self, filename):
-        """Check the extension of a resource."""
-        for file_ in ["test/grant_authorization_button.png", filename]:
-            id_ = self.api.add_resource(filename=file_)
-            resource = self.api.get_resource(id_=id_, fields="file_extension")
-            self.assertEqual(resource["file_extension"], os.path.splitext(file_)[1][1:])
+    def test_check_property_title(self, filename: str):
+        """Check the title of a resource."""
+        title = self.get_random_string()
+        id_ = self.api.add_resource(filename=filename, title=title)
+        resource = self.api.get_resource(id_=id_)
+        self.assertEqual(resource["title"], title)
 
 
 # TODO: Add more tests for the search parameter.
@@ -559,41 +577,38 @@ class Search(TestBase):
         self.assertEqual(self.api.search(query="*"), self.empty_search)
 
     def test_notes(self):
-        """Search by notebooks and search endpoint should yield same results."""
-        self.skipTest("TODO: Seems to be broken.")
+        """
+        Wildcard search for all notes is disabled, because of performance reasons.
+        See: https://github.com/laurent22/joplin/issues/5546
+        """
         self.api.add_notebook()
         self.api.add_note()
-        self.assertEqual(
-            self.api.search(query="*"), self.api.get_notes(),
-        )
-        self.assertEqual(
-            self.api.search(query="*", type="note"), self.api.get_notes(),
-        )
+        self.assertEqual(self.api.search(query="*"), self.empty_search)
+        self.assertEqual(self.api.search(query="*", type="note"), self.empty_search)
 
     def test_notebooks(self):
         """Search by notebooks and search endpoint should yield same results."""
         self.api.add_notebook()
         self.assertEqual(
-            self.api.search(query="*", type="folder"), self.api.get_notebooks(),
+            self.api.search(query="*", type="folder"),
+            self.api.get_notebooks(),
         )
 
-    def test_sub_notebooks(self):
-        """Search for all notebooks, contained by a specific notebook."""
-        self.skipTest("TODO: Not possible yet?")
-
     @with_resource
     def test_resources(self, filename):
         """Search by resources and search endpoint should yield same results."""
         self.api.add_resource(filename=filename)
         self.assertEqual(
-            self.api.search(query="*", type="resource"), self.api.get_resources(),
+            self.api.search(query="*", type="resource"),
+            self.api.get_resources(),
         )
 
     def test_tags(self):
         """Search by tags and search endpoint should yield same results."""
         self.api.add_tag()
         self.assertEqual(
-            self.api.search(query="*", type="tag"), self.api.get_tags(),
+            self.api.search(query="*", type="tag"),
+            self.api.get_tags(),
         )
 
     def test_master_key(self):
@@ -720,6 +735,17 @@ class Tag(TestBase):
         for tag in tags["items"]:
             self.assertEqual(list(tag.keys()), self.default_properties)
 
+    def test_get_all_tags(self):
+        """Get all tags, unpaginated."""
+        # Small limit and count to create/remove as less as possible items.
+        count, limit = random.randint(1, 10), random.randint(1, 10)
+        for _ in range(count):
+            self.api.add_tag()
+        self.assertEqual(
+            len(self.api.get_tags(limit=limit)["items"]), min(limit, count)
+        )
+        self.assertEqual(len(self.api.get_all_tags(limit=limit)), count)
+
     def test_get_tags_valid_properties(self):
         """Try to get specific properties of a tag."""
         self.api.add_tag()
@@ -738,7 +764,10 @@ class Fuzz(TestBase):
             method = random.choice(("delete", "get", "post", "put"))
             try:
                 self.api._request(method, path)
-            except requests.exceptions.HTTPError:
+            except (
+                requests.exceptions.HTTPError,
+                requests.packages.urllib3.exceptions.LocationParseError,
+            ):
                 pass
         self.api.ping()
 
@@ -785,6 +814,19 @@ class Miscellaneous(TestBase):
 
 
 class Regression(TestBase):
+    @unittest.skip("Enable when the bug is fixed.")
+    @unittest.skipIf(not SLOW_TESTS, "Generating the long string is slow.")
+    def test_long_body(self):
+        """
+        https://github.com/laurent22/joplin/issues/5543
+        Response HTTP 104: https://stackoverflow.com/a/52826181
+        """
+        # Use only one random character, since it's very slow already.
+        body = self.get_random_string(1) * 10 ** 9
+        self.api.add_notebook()
+        note_id = self.api.add_note(body=body)
+        self.assertEqual(self.api.get_note(id_=note_id)["title"], body)
+
     def test_note_tag_fields(self):
         """https://github.com/laurent22/joplin/issues/4407"""
         self.api.add_notebook()
